@@ -15,6 +15,7 @@
 #
 
 from __future__ import annotations
+
 import doctest
 import inspect
 import os
@@ -36,6 +37,8 @@ author = "Mahmoud Asem"
 # -- General configuration ---------------------------------------------------
 master_doc = "index"
 
+html_static_path = ["_static"]
+html_logo = "_static/logo.svg"
 
 # Add any Sphinx extension module names here, as strings. They can be
 # extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
@@ -48,7 +51,7 @@ extensions = [
     "sphinx.ext.intersphinx",
     "sphinx.ext.linkcode",
     "sphinx.ext.napoleon",
-    "sphinxcontrib.bibtex",
+    # "sphinxcontrib.bibtex",
     "sphinxcontrib.katex",
     "sphinx_autodoc_typehints",
     "nbsphinx",
@@ -88,6 +91,13 @@ html_theme = "sphinx_book_theme"
 # html_static_path = ["_static"]
 
 
+html_theme_options = {
+    "show_toc_level": 2,
+    "repository_url": "https://github.com/ASEM000/PyTreeClass",
+    "use_repository_button": True,
+}
+
+
 # -- Options for doctest -----------------------------------------------------
 
 doctest_test_doctest_blocks = "true"
@@ -151,7 +161,7 @@ def linkcode_resolve(domain, info):
     except OSError:
         return None
 
-    return "https://github.com/ASEM000/pytreeclass/%s#L%d#L%d" % (
+    return "https://github.com/ASEM000/pytreeclass/blob/main/pytreeclass/%s#L%d#L%d" % (
         os.path.relpath(filename, start=os.path.dirname(pytreeclass.__file__)),
         lineno,
         lineno + len(source) - 1,
@@ -164,12 +174,3 @@ nbsphinx_execute = "never"
 nbsphinx_codecell_lexer = "ipython"
 nbsphinx_kernel_name = "python"
 nbsphinx_timeout = 180
-nbsphinx_prolog = r"""
-{% set docname = 'docs/' + env.doc2path(env.docname, base=None) %}
-.. only:: html
-    .. role:: raw-html(raw)
-        :format: html
-    .. nbinfo::
-        Interactive online version:
-        :raw-html:`<a href="https://colab.research.google.com/github/ASEM000/pytreeclass/blob/main/{{ docname }}"><img alt="Open In Colab" src="https://colab.research.google.com/assets/colab-badge.svg" style="vertical-align:text-bottom"></a>`
-"""
\ No newline at end of file
