@@ -10,8 +10,10 @@ from pytreeclass._src.tree_freeze import (
 from pytreeclass._src.tree_indexer import bcmap, tree_indexer
 from pytreeclass._src.tree_pprint import (
     tree_diagram,
+    tree_indent,
     tree_mermaid,
     tree_repr,
+    tree_repr_with_trace,
     tree_str,
     tree_summary,
 )
@@ -34,7 +36,9 @@ __all__ = (
     "tree_mermaid",
     "tree_repr",
     "tree_str",
+    "tree_indent",
     "tree_summary",
+    "tree_trace_summary",
     # freeze/unfreeze utils
     "is_nondiff",
     "is_frozen",
@@ -48,6 +52,7 @@ __all__ = (
     "tree_map_with_trace",
     "tree_leaves_with_trace",
     "tree_flatten_with_trace",
+    "tree_repr_with_trace",
 )
 
-__version__ = "0.2.2"
+__version__ = "0.2.4"
