@@ -4,21 +4,23 @@
 from __future__ import annotations
 
 import functools as ft
+import operator as op
 from collections.abc import Callable
-from typing import Any, NamedTuple
+from math import ceil, floor, trunc
+from typing import Any, NamedTuple, TypeVar
 
 import jax
 import jax.numpy as jnp
 import jax.tree_util as jtu
 import numpy as np
 
-from pytreeclass._src.tree_freeze import _call_context
+from pytreeclass._src.tree_decorator import _mutable_context
 from pytreeclass._src.tree_trace import tree_map_with_trace
 
 PyTree = Any
-EllipsisType = type(Ellipsis)
 _no_initializer = object()
 _non_partial = object()
+T = TypeVar("T")
 
 
 def _is_leaf_bool(node: Any) -> bool:
@@ -33,6 +35,10 @@ def _check_valid_mask_leaf(where: Any):
     return where
 
 
+def _true_tree(tree: PyTree, is_leaf) -> PyTree:
+    return jtu.tree_map(lambda _: True, tree, is_leaf=is_leaf)
+
+
 # tree indexing by boolean PyTree
 
 
@@ -132,12 +138,14 @@ class _TreeAtMask(NamedTuple):
     where: PyTree
 
     def get(self, *, is_leaf: Callable[[Any], bool] | None = None) -> PyTree:
-        return _get_at_mask(self.tree, self.where, is_leaf)
+        where = _true_tree(self.tree, is_leaf) if self.where is ... else self.where
+        return _get_at_mask(self.tree, where, is_leaf)
 
     def set(
         self, set_value: Any, *, is_leaf: Callable[[Any], bool] | None = None
     ) -> PyTree:
-        return _set_at_mask(self.tree, self.where, set_value, is_leaf)
+        where = _true_tree(self.tree, is_leaf) if self.where is ... else self.where
+        return _set_at_mask(self.tree, where, set_value, is_leaf)
 
     def apply(
         self,
@@ -145,7 +153,8 @@ class _TreeAtMask(NamedTuple):
         *,
         is_leaf: Callable[[Any], bool] | None = None,
     ):
-        return _apply_at_mask(self.tree, self.where, func, is_leaf)
+        where = _true_tree(self.tree, is_leaf) if self.where is ... else self.where
+        return _apply_at_mask(self.tree, where, func, is_leaf)
 
     def reduce(
         self,
@@ -154,7 +163,8 @@ class _TreeAtMask(NamedTuple):
         *,
         is_leaf: Callable[[Any], bool] | None = None,
     ) -> Any:
-        return _reduce_at_mask(self.tree, self.where, func, initializer, is_leaf)
+        where = _true_tree(self.tree, is_leaf) if self.where is ... else self.where
+        return _reduce_at_mask(self.tree, where, func, initializer, is_leaf)
 
 
 def _at_mask(tree: PyTree, where: PyTree) -> PyTree:
@@ -197,8 +207,8 @@ def _get_at_trace(
         names, _, indices, __ = trace
         for i, item in enumerate(where):
             try:
-                if (isinstance(item, int) and indices[i + 1] != item) or (
-                    isinstance(item, str) and names[i + 1] != item
+                if (isinstance(item, int) and indices[i] != item) or (
+                    isinstance(item, str) and names[i] != item
                 ):
                     return None
             except IndexError:
@@ -224,8 +234,8 @@ def _set_at_trace(
         names, _, indices, __ = trace
         for i, item in enumerate(where):
             try:
-                if (isinstance(item, int) and indices[i + 1] != item) or (
-                    isinstance(item, str) and names[i + 1] != item
+                if (isinstance(item, int) and indices[i] != item) or (
+                    isinstance(item, str) and names[i] != item
                 ):
                     return leaf
             except IndexError:
@@ -256,8 +266,8 @@ def _apply_at_trace(
 
         for i, item in enumerate(where):
             try:
-                if (isinstance(item, int) and indices[i + 1] != item) or (
-                    isinstance(item, str) and names[i + 1] != item
+                if (isinstance(item, int) and indices[i] != item) or (
+                    isinstance(item, str) and names[i] != item
                 ):
                     return leaf
             except IndexError:
@@ -288,8 +298,8 @@ def _reduce_at_trace(
 
         for i, item in enumerate(where):
             try:
-                if (isinstance(item, int) and indices[i + 1] != item) or (
-                    isinstance(item, str) and names[i + 1] != item
+                if (isinstance(item, int) and indices[i] != item) or (
+                    isinstance(item, str) and names[i] != item
                 ):
                     return None
             except IndexError:
@@ -340,7 +350,7 @@ class _TreeAtPath(NamedTuple):
         # this method first creates a copy of the tree,
         # next it unfreezes the tree then calls the method on the attribute
         # and finally freezes the tree again
-        with _call_context(self.tree) as tree:
+        with _mutable_context(self.tree, kopy=True) as tree:
             # check if the attribute is a string
             (method_name,) = self.where
             if not isinstance(method_name, str):
@@ -430,14 +440,11 @@ def tree_indexer(tree: PyTree) -> PyTree:
                 # `register_pytree_node_trace_func`
                 return _at_trace(tree=tree, where=(where,))
 
-            if isinstance(where, type(tree)):
+            if isinstance(where, (type(tree), type(...))):
                 # indexing by boolean pytree
-                return _at_mask(tree=tree, where=where)
-
-            if isinstance(where, EllipsisType):
                 # Ellipsis as an alias for all elements
                 # model.at[model == model ] <--> model.at[...]
-                return tree.at[tree == tree]
+                return _at_mask(tree=tree, where=where)
 
             raise NotImplementedError(
                 f"Indexing with {type(where).__name__} is not implemented.\n"
@@ -512,8 +519,8 @@ def bcmap(
 
         >>> # using **numpy** functions on pytrees
         >>> import jax.numpy as jnp
-        >>> pytc.bcmap(jnp.add)([1,2,3],[1,2,3])
-        [Array(2, dtype=int32, weak_type=True), Array(4, dtype=int32, weak_type=True), Array(6, dtype=int32, weak_type=True)]
+        >>> pytc.bcmap(jnp.add)([1,2,3],[1,2,3]) # doctest: +SKIP
+        [2, 4, 6]
     """
 
     @ft.wraps(func)
@@ -568,3 +575,120 @@ def bcmap(
     docs = f"Broadcasted version of {func.__name__}\n{func.__doc__}"
     wrapper.__doc__ = docs
     return wrapper
+
+
+def _unary_op(func):
+    def wrapper(self):
+        return jtu.tree_map(func, self)
+
+    return ft.wraps(func)(wrapper)
+
+
+def _binary_op(func):
+    def wrapper(lhs, rhs=None):
+        if isinstance(rhs, type(lhs)):
+            return jtu.tree_map(func, lhs, rhs)
+        return jtu.tree_map(lambda x: func(x, rhs), lhs)
+
+    return ft.wraps(func)(wrapper)
+
+
+def _swop(func):
+    # swaping the arguments of a two-arg function
+    return ft.wraps(func)(lambda lhs, rhs: func(rhs, lhs))
+
+
+def _leafwise_transform(klass: type[T]) -> type[T]:
+    # add leafwise transform methods to the class
+    # that enable the user to apply a function to
+    # all the leaves of the tree
+    for key, method in (
+        ("__abs__", _unary_op(abs)),
+        ("__add__", _binary_op(op.add)),
+        ("__and__", _binary_op(op.and_)),
+        ("__ceil__", _unary_op(ceil)),
+        ("__divmod__", _binary_op(divmod)),
+        ("__eq__", _binary_op(op.eq)),
+        ("__floor__", _unary_op(floor)),
+        ("__floordiv__", _binary_op(op.floordiv)),
+        ("__ge__", _binary_op(op.ge)),
+        ("__gt__", _binary_op(op.gt)),
+        ("__invert__", _unary_op(op.invert)),
+        ("__le__", _binary_op(op.le)),
+        ("__lshift__", _binary_op(op.lshift)),
+        ("__lt__", _binary_op(op.lt)),
+        ("__matmul__", _binary_op(op.matmul)),
+        ("__mod__", _binary_op(op.mod)),
+        ("__mul__", _binary_op(op.mul)),
+        ("__ne__", _binary_op(op.ne)),
+        ("__neg__", _unary_op(op.neg)),
+        ("__or__", _binary_op(op.or_)),
+        ("__pos__", _unary_op(op.pos)),
+        ("__pow__", _binary_op(op.pow)),
+        ("__radd__", _binary_op(_swop(op.add))),
+        ("__rand__", _binary_op(_swop(op.and_))),
+        ("__rdivmod__", _binary_op(_swop(divmod))),
+        ("__rfloordiv__", _binary_op(_swop(op.floordiv))),
+        ("__rlshift__", _binary_op(_swop(op.lshift))),
+        ("__rmatmul__", _binary_op(_swop(op.matmul))),
+        ("__rmod__", _binary_op(_swop(op.mod))),
+        ("__rmul__", _binary_op(_swop(op.mul))),
+        ("__ror__", _binary_op(_swop(op.or_))),
+        ("__round__", _binary_op(round)),
+        ("__rpow__", _binary_op(_swop(op.pow))),
+        ("__rrshift__", _binary_op(_swop(op.rshift))),
+        ("__rshift__", _binary_op(op.rshift)),
+        ("__rsub__", _binary_op(_swop(op.sub))),
+        ("__rtruediv__", _binary_op(_swop(op.truediv))),
+        ("__rxor__", _binary_op(_swop(op.xor))),
+        ("__sub__", _binary_op(op.sub)),
+        ("__truediv__", _binary_op(op.truediv)),
+        ("__trunc__", _unary_op(trunc)),
+        ("__xor__", _binary_op(op.xor)),
+    ):
+        if key not in vars(klass):
+            # do not override any user defined methods
+            # this behavior similar is to `dataclasses.dataclass`
+            setattr(klass, key, method)
+    return klass
+
+
+def _is_lhs_rhs_equal(lhs, rhs) -> bool | jax.Array:
+    if hasattr(lhs, "shape") and hasattr(lhs, "dtype"):
+        if hasattr(rhs, "shape") and hasattr(rhs, "dtype"):
+            verdict = jnp.array_equal(lhs, rhs)
+            try:
+                return bool(verdict)
+            except Exception:
+                return verdict  # fail under `jit`
+        return False
+    return lhs == rhs
+
+
+def is_tree_equal(*trees: Any) -> bool | jax.Array:
+    """Return `True` if all pytrees are equal.
+
+    Note:
+        trees are compared using their leaves and treedefs.
+        For `array` leaves `jnp.array_equal` is used, for other leaves
+        method `__eq__` is used.
+
+    Note:
+        Under `jit` the return type is boolean `jax.Array` instead of python `bool`.
+    """
+
+    tree0, *rest = trees
+    leaves0, treedef0 = jtu.tree_flatten(tree0)
+    verdict = True
+
+    for tree in rest:
+        leaves, treedef = jtu.tree_flatten(tree)
+        if (treedef != treedef0) or verdict is False:
+            return False
+        verdict = ft.reduce(op.and_, map(_is_lhs_rhs_equal, leaves0, leaves), verdict)
+    return verdict
+
+
+def tree_copy(tree: PyTree) -> PyTree:
+    """Return a copy of the tree."""
+    return jtu.tree_unflatten(*jtu.tree_flatten(tree)[::-1])  # type: ignore
